package com.unitesk.requality.eclipse.tools;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;

public class DeltaProcessor implements IResourceChangeListener {
    private Map<IFile, UUID> resourceToNode;
    private ResourceStorage rs;
    private Map<UUID, IFile> added = new LinkedHashMap<UUID, IFile>();
    private Map<UUID, IFile> deleted = new LinkedHashMap<UUID, IFile>();
    private Map<UUID, IFile[]> moved = new LinkedHashMap<UUID, IFile[]>();

    public DeltaProcessor(Map<IFile, UUID> resourceToNode, ResourceStorage rs) {
        this.resourceToNode = resourceToNode;
        this.rs = rs;
    }

    @Override
    public void resourceChanged(IResourceChangeEvent event) {
        if (event.getType() == IResourceChangeEvent.PRE_DELETE
                || event.getType() == IResourceChangeEvent.PRE_CLOSE) {
            return;
        }
        if (event.getDelta().getResource().getType() == IResource.ROOT) {
            for (IResourceDelta rd : event.getDelta().getAffectedChildren()) {
                if (rd.getResource().getProject().equals(rs.getProject())) {
                    processDelta(rd);
                }
            }
        } else {
            throw new IllegalArgumentException(
                    "Delta resource must be root type");
        }
    }

    public void processDelta(IResourceDelta delta) {
        moved.clear();
        deleted.clear();
        added.clear();

        Map<IFile, Integer> result = collectChanges(delta, null);
        for (IFile f : result.keySet()) {
            if (result.get(f) == IResourceDelta.REMOVED) {
                processChange(null, f);
            } else if(result.get(f) == IResourceDelta.ADDED) {
                processChange(f, null);
            } else {
                processChange(f, f);
            }
        }

        // find move action
        for(UUID uuid: deleted.keySet()){
            if(added.get(uuid)!=null
                    // TODO: Track as move not only rename
                    && added.get(uuid).getParent().equals(deleted.get(uuid).getParent())) {
                moved.put(uuid, new IFile[] { added.get(uuid), deleted.get(uuid)});
            }
        }
        for(UUID uuid:moved.keySet()) {
            added.remove(uuid);
            deleted.remove(uuid);
        }


        // apply changes to cache
        for(UUID uuid: deleted.keySet()) {
            rs.getDeltaTarget().deleted(uuid);
            resourceToNode.remove(deleted.get(uuid));
        }

        for(UUID uuid: added.keySet()) {
            resourceToNode.put(added.get(uuid), uuid);
            rs.getDeltaTarget().created(uuid);
        }

        for(UUID uuid: moved.keySet()) {
            IFile to = moved.get(uuid)[0];
            IFile from = moved.get(uuid)[1];
            if(from.getParent().equals(to.getParent())) { // rename move
                resourceToNode.remove(from);
                resourceToNode.put(to, uuid);
                rs.getDeltaTarget().changed(uuid);
            }
            // TODO: complex move processing
//            else {
//                String opId = from.getParent().getName();
//                IFile oldParentFile = from.getParent().getParent().getFile(new Path(opId+".json"));
//                UUID fromParent = getUUIdFromMoved(oldParentFile);
//                if(fromParent==null) {
//                    fromParent = resourceToNode.get(oldParentFile);
//                }
//                resourceToNode.put(to, uuid);
//                resourceToNode.remove(from);
//                rs.getDeltaTarget().moved(uuid, fromParent, rs.getParent(to));
//            }
        }
    }

    public UUID getUUIdFromMoved(IFile file) {
        for(UUID uuid: moved.keySet()) {
            if(moved.get(uuid)[0].equals(file)) {
                return uuid;
            }
            if(moved.get(uuid)[1].equals(file)) {
                return uuid;
            }
        }
        return null;
    }
    
    private boolean isNode(IResource resource) {
        if (resourceToNode.containsKey(resource)) {
            return true;
        }
        if (resource.getType() == IResource.FILE) {
            IFile file = (IFile) resource;
            if (file.getFileExtension() != null
                    && file.getFileExtension().equals("json")) {
                return true;
            }
        }
        return false;
    }

    private void processChange(IFile fileAdd, IFile fileDel) {
        // Added
        if (fileAdd != null && fileDel == null) {
            UUID uuid = rs.getUUId(fileAdd);
            added.put(uuid, fileAdd);
        }
        // Removed
        else if (fileAdd == null && fileDel != null) {
            UUID uuid = resourceToNode.get(fileDel);
            deleted.put(uuid, fileDel);
        }
        // Changed
        else if (fileAdd != null && fileAdd == fileDel) {
            UUID uuid = rs.getUUId(fileAdd);
            rs.getDeltaTarget().changed(uuid);
        } else {
            // UUID uuid = getUUId(fileAdd);
            // resourceToNode.put(fileAdd, uuid);
            // if (fileAdd.getParent().equals(fileDel.getParent())) {
            // resourceToNode.remove(fileDel);
            // deltaTarget.changed(uuid);
            // } else {
            // resourceToNode.remove(fileDel);
            // deltaTarget.moved(uuid, parentUUId);
            // }
        }
    }

    private Map<IFile, Integer> collectChanges(IResourceDelta delta,
            Map<IFile, Integer> changes) {
        Map<IFile, Integer> result = changes;
        if (result == null) {
            result = new LinkedHashMap<IFile, Integer>();
        }

        IResourceDelta subDeltas[] = delta.getAffectedChildren();

        for (IResourceDelta sub : subDeltas) {
            IResource res = sub.getResource();
            if ((sub.getKind() & IResourceDelta.ADDED) != 0) {
                if (isNode(res)) {
                    result.put((IFile) res, IResourceDelta.ADDED);
                }
            }

            if ((sub.getKind() & IResourceDelta.CHANGED) != 0) {
                if (isNode(res)) {
                    result.put((IFile) res, IResourceDelta.CHANGED);
                }
            }

            collectChanges(sub, result);

            if ((sub.getKind() & IResourceDelta.REMOVED) !=0) {
                if (isNode(res)) {
                    result.put((IFile) res, IResourceDelta.REMOVED);
                }
            }
        }
        return result;
    }
}
