package com.unitesk.requality.merge;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class Context {

    private long upperBound;
    private long lowerBound;
    private long width;
    private long label;
    private int id;
    private int sibCount;
    private int position;
    private int depth;
    private int index;
    private Context parentContext;
    private List<Context> childContexts;
    private Node associatedNode;
    private int weight=1;

    public Context(Node root) {
	associatedNode = root;
	upperBound = Long.MAX_VALUE;
	lowerBound = 0;
	width = upperBound - lowerBound;
	label = makeLabel();
	id = 1;
	sibCount = 0;
	position = 1;
	depth = 0;
	index = 1;
	parentContext = null;
	childContexts = new ArrayList<Context>();
    }

    public Context(Node node, Context parentContext) {
	associatedNode = node;
	sibCount = node.getDomNode().getParentNode().getChildNodes()
		.getLength();
	position = 1;
	for (org.w3c.dom.Node x = node.getDomNode(); x.getPreviousSibling() != null; x = x
		.getPreviousSibling())
	    position++;
	width = parentContext.width / sibCount;
	upperBound = parentContext.lowerBound + width * position;
	lowerBound = upperBound - width;
	label = makeLabel();
	id = parentContext.id + position;
	this.parentContext = parentContext;
	depth = parentContext.depth + 1;
	childContexts = new ArrayList<Context>();
	parentContext.addChildContext(this);
	index = 1;
	for (int i = 0; i < parentContext.childContexts.indexOf(this)
		&& i < parentContext.childContexts.size(); i++)
	    if (parentContext.childContexts.get(i).associatedNode.getDomNode()
		    .getNodeName()
		    .equals(this.associatedNode.getDomNode().getNodeName()))
		this.index++;
    }
    
    private long makeLabel() {
	Random rnd = new Random(System.nanoTime());
	long label = lowerBound+width/2;
	label += (long) (rnd.nextDouble()*width/4);
	return label;
    }
    
    private void addChildContext(Context context) {
	if (childContexts.isEmpty())
	    childContexts.add(context);
	else if (childContexts.get(childContexts.size() - 1).label < context.label)
	    childContexts.add(context);
	else {
	    for (int i = 0; i < childContexts.size(); i++) {
		Context prev = childContexts.get(i);
		if (prev.label > context.label)
		    childContexts.add(i, context);
	    }

	}
    }

    public List<Context> getChildContexts() {
	return childContexts;
    }

    public List<Context> getSiblingContexts() {
	if (parentContext == null)
	    return new ArrayList<Context>();
	else
	    return parentContext.childContexts;
    }

    private List<Context> getNeighborhood() {
	List<Context> neighbors = getFullNeighborhood();
	neighbors.remove(this);
	return neighbors;
    }

    private List<Context> getFullNeighborhood() {
	List<Context> neighbors = new ArrayList<Context>();
	if (parentContext != null) {
	    neighbors.add(parentContext);
	    neighbors.addAll(parentContext.getSiblingContexts());
	}
	neighbors.addAll(this.getSiblingContexts());
	neighbors.addAll(this.getChildContexts());
	return neighbors;
    }

    private List<Context> getTextLeaves() {
	List<Context> textLeaves = new ArrayList<Context>();
	for(Context c1 : this.childContexts) {
	    if(c1.associatedNode.isText())
		textLeaves.add(c1);
	    else if(!c1.associatedNode.isLeaf())
		textLeaves.addAll(c1.getTextLeaves());
	}
	return textLeaves;
    }
    
    private int neighborhoodDistance(Context that) {
	List<Context> thisNeighbors = this.getNeighborhood();
	List<Context> thatNeighbors = that.getNeighborhood();
	int distance = 0;
	for(int i=0; i<Math.min(thisNeighbors.size(), thatNeighbors.size()); i++) {
	    Context x = thisNeighbors.get(i);
	    Context y = thatNeighbors.get(i);
	    distance += x.contextDistance(y);
	}
	return distance;
    }

    private int contextDistance(Context that) {
	int distance = 0;
	distance += Math.abs(this.id-that.id);
	distance += Math.abs(this.position-that.position);
	distance += Math.abs(this.depth-that.depth);
	distance += Math.abs(this.sibCount-that.sibCount);
	distance += Math.abs(this.index-that.index);
	if(this.parentContext!=null && that.parentContext!=null)
	    distance += this.parentContext.contextDistance(that.parentContext);
	if(this.associatedNode.isText() && that.associatedNode.isText())
	    distance += textDistance(that);
	return distance;
    }
    
    private int textDistance(Context that) {
	String a = this.associatedNode.getDomNode().getNodeValue();
	String b = that.associatedNode.getDomNode().getNodeValue();
	int distance = Math.abs(a.length() - b.length());
	for (int i = 0; i < Math.min(a.length(), b.length()); i++) {
	    if (a.charAt(i) != b.charAt(i))
		distance++;
	}
	return distance;
    }
    
    private int labelCoeff(Context that) {
	return (int) (Math.abs(this.label-that.label)/this.width);
    }
    
    public int contextualDistance(Context that) {
	if(!this.associatedNode.isLeaf() && !that.associatedNode.isLeaf()) {
	    List<Context> thisTextLeaves = this.getTextLeaves();
	    List<Context> thatTextLeaves = that.getTextLeaves();
	    if(!thisTextLeaves.isEmpty() && !thatTextLeaves.isEmpty()) {
		int distance = 0;
		for(Context thisLeaf : thisTextLeaves) {
		    int min_distance = Integer.MAX_VALUE;
		    for(Context thatLeaf : thatTextLeaves) {
			int currentDistance = thisLeaf.textDistance(thatLeaf);
			if(currentDistance<min_distance)
			    min_distance = currentDistance;
		    }
		    distance += min_distance;
		}
		return distance;
	    }
	}
	return Integer.MAX_VALUE;
    }
    
    public boolean equals(Context that) {
	if(this.label==that.label)
	    return true;
	else
	    return false;
    }
    
    public int positionalDistance(Context that) {
	int distance = (this.contextDistance(that)+this.neighborhoodDistance(that))*this.labelCoeff(that);
	distance *= Math.max(this.weight, that.weight);
	return distance;
    }
    
    public int getPosition() {
	return position;
    }

    public int getDepth() {
	return depth;
    }

    public boolean isWithin(Context context) {
	if (this.lowerBound >= context.lowerBound
		&& this.upperBound <= context.upperBound)
	    return true;
	return false;
    }

    public Context getParentContext() {
	return parentContext;
    }

    public long getLabel() {
	return label;
    }

    public String toString() {
	return associatedNode.toString();
    }
    
    public void setWeight(int value) {
	this.weight = value;
    }
    
    public int getIndex() {
	return this.index;
    }
    
    public int getId() {
	return this.id;
    }
}
