/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.report;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import java.util.UUID;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.events.Characters;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.nodetypes.Requirement;

import freemarker.template.Configuration;
import freemarker.template.DefaultObjectWrapper;
import freemarker.template.Template;

/**
 * This class is introduced as <i><b>this</b></i> var and all public methods can
 * be used in freemarker template.
 */
public class ReportGenerator {

    private static QName TEMPLATE = new QName("template");
    private static QName FILE = new QName("file");
    private static QName DIR = new QName("dir");
    private static QName SOURCE = new QName("source");
    private static QName LOCATION = new QName("location");
    private static String DATA_OBJECT = "object";
    private static String DATA_NAME = "name";
    private static String DATA_VALUE = "value";

    public ReportGenerator() {
    }

    public static ReportGenerator newInstance() {
        return new ReportGenerator();
    }

    public String generateReport(TreeDB db, String[] args) {
        String templatePath = args[0];
        String dataPath = args[1];
        String catPath = args[2];
        String rootId = args[3];
        String locationDir = args[4];

        // DocParser docParser = new DocParser();
        // String ftlPath = docParser.parseTemplate( templatePath );
        if (rootId.equals("/")) {
            rootId = "";
        }
        try {
            return generateReport(templatePath, dataPath, catPath, rootId,
                    locationDir, db);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        return "";
    }

    /**
     * This method is called from FReqTools.
     * 
     * @param templatePath
     *            .
     * @param dataPath
     *            .
     * @param catPath
     *            .
     * @param rootId
     *            .
     * @param saveTo
     *            .
     * @return .
     * @throws FileNotFoundException .
     */
    private String generateReport(String templatePath, String dataPath,
            String catPath, String rootId, String saveTo, TreeDB db)
            throws FileNotFoundException {
        if (!new File(templatePath).exists()) {
            System.err.println("Wrong path to template.xml: " + templatePath);
            return "";
        }
        if (!new File(dataPath).exists()) {
            System.err.println("Wrong path to data: " + dataPath);
            return "";
        }
        if (!new File(catPath).exists()) {
            System.err.println("Wrong path to catalogue: " + catPath);
            return "";
        }
        if (!new File(saveTo).exists()) {
            System.out.println("Creating location path: " + saveTo);
            File f = new File(saveTo);
            boolean result = f.mkdirs();
            if (!result) {
                System.err.println("Couldn't create location dir");
                return "";
            }
        }

        db.getNode(Requirement.getTypeRootQId());

        TreeNode root;
        if (rootId.length() > 0) {
            root = db.getNode(UUID.fromString(rootId));
        } else {
            root = db.getNode(Requirement.getTypeRootQId());
        }
        if (root == null) {
            System.out.println("Root requirement is null");
            throw new RuntimeException("Root requirement is null");
        }

        TreeDB instTree = new TreeDB(new TreeInstancer(db));
//        NodeInstance instancse = new NodeInstance(root,
//                Collections.<String, String> emptyMap());
        try {
            return generateReport(instTree, templatePath, processData(dataPath), saveTo);
        } catch (RuntimeException e) {
            e.printStackTrace();
        }
        return "";
    }

    /**
     * Generates report.
     * 
     * @param instTree
     *            root requirement
     * @param template
     *            xml mapping file.
     * @param data
     *            xml file with additional information for report. Must contains
     *            'object' tags with attributes <i>name</i> and <i>value</i>
     * @param location
     *            path to file system catalogue where report should be saved.
     * @return path to file system catalogue with generated report.
     * @throws java.io.FileNotFoundException .
     */
    public String generateReport(TreeDB instTree, String template,
            Map<String, Object> data, String location) throws FileNotFoundException {
        String result = "";
        if (instTree != null && template != null && template.length() > 0) {
            Map<String, Object> vars = new HashMap<String, Object>(data);
            String rootReqQId = (String)data.get("rootReqQId");
            vars.put("root", instTree.getRootNode());
            // vars.put( "map", map );
            vars.put("this", this);
            vars.put("vars", vars);
            vars.put("source", new File(template).getAbsoluteFile().getParent());
            vars.put("rootReq", instTree.getNode(rootReqQId));
            vars.put("reportLocation", location);

            result = processTemplates(template, vars,
                    location != null ? location : "");
            log("Report generation completed.");
        } else {
            log("Some of parameters aren't defined!");
        }
        return result;
    }

    private String processTemplates(String source, Map<String, Object> vars,
            String location) {
        String first = "";
        File file = new File(source);
        File saveTo = location.length() > 0 ? new File(location)
                : new File(".");
        saveTo.mkdirs();
        if (file.exists() && file.isFile()) {
            try {
                XMLInputFactory factory = XMLInputFactory.newInstance();
                Reader fReader = new InputStreamReader( new FileInputStream(file.getPath()),"UTF-8" );
                XMLEventReader eReader = factory.createXMLEventReader(fReader);
                while (eReader.hasNext()) {

                    String res;
                    XMLEvent event = eReader.nextEvent();
                    switch (event.getEventType()) {
                    case XMLStreamConstants.START_ELEMENT:
                        res = processElem(event.asStartElement(), vars, file
                                .getAbsoluteFile().getParentFile(), saveTo);
                        if (first.length() == 0) {
                            first = res;
                        }
                        break;
                    case XMLStreamConstants.END_ELEMENT:
                    case XMLStreamConstants.CHARACTERS:
                        break;
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            log("File " + file.getPath() + " does not exist.");
        }
        return first;
    }

    private String processElem(StartElement elem, Map<String, Object> vars,
            File folder, File saveTo) {
        javax.xml.stream.events.Attribute attr;
        attr = elem.getAttributeByName(SOURCE);
        String source = attr != null ? attr.getValue() : "";
        attr = elem.getAttributeByName(LOCATION);
        String location = attr != null ? new File(saveTo, attr.getValue())
                .getPath() : "";

        QName name = elem.getName();
        source = new File(folder, source).getAbsolutePath();
        if (name.equals(TEMPLATE)) {
            processTemplate(source, vars, location);
        } else if (name.equals(FILE)) {
            copyFile(source, location);
        } else if (name.equals(DIR)) {
            copyDir(source, location);
        }
        return location;
    }

    /**
     * Copies catalogue into selected place.
     * 
     * @param source
     *            path to catalogue which must be copied.
     * @param location
     *            path to where catalogue must be copied.
     * @return true if coping was successful, false - otherwise.
     */
    public boolean copyDir(String source, String location) {
        String fileSeparator = System.getProperty("file.separator");
        source = source.replace("/", fileSeparator);
        location = location.replace("\\", fileSeparator);
        boolean result = true;
        File loc = new File(location);
        if (!loc.exists()) {
            loc.mkdirs();
        }
        File dir = new File(source);
        if (dir.exists() && dir.isDirectory()) {
            File[] files = dir.listFiles();
            for (File file : files) {
                String filePath = file.getPath();
                String locPath = new File(location, file.getName()).getPath();

                if (file.isDirectory()) {
                    copyDir(filePath, locPath);
                } else {
                    copyFile(filePath, locPath);
                }
            }
        } else {
            log("Directory " + dir.getPath() + " does not exist.");
            result = false;
        }
        return result;
    }

    /**
     * Copies file into selected place.
     * 
     * @param source
     *            path to file which must be copied.
     * @param location
     *            path to location where file must be copied.
     * @return true if coping successful, false - otherwise.
     */
    public boolean copyFile(String source, String location) {
        boolean result = true;

        try {
            String fileSeparator = System.getProperty("file.separator");
            source = source.replace("/", fileSeparator);
            location = location.replace("\\", fileSeparator);
            File inFile = new File(source);
            if (inFile.exists() && inFile.isFile()) {
                File outFile = new File(location);
                if (outFile.exists()) {
                    log("File " + outFile.getPath() + " already exists.");
                    outFile.delete();
                }
                if (!outFile.exists()) {
                    System.out.println("Creating input stream from file: "
                            + inFile.getPath());

                    InputStream in = new FileInputStream(inFile);
                    OutputStream out = new FileOutputStream(outFile);
                    byte[] buf = new byte[1024];
                    int len;
                    while ((len = in.read(buf)) > 0) {
                        out.write(buf, 0, len);
                    }
                    in.close();
                    out.close();
                } else {
                    log("Cannot delete file " + outFile.getPath() + ".");
                }
            } else {
                log("File " + inFile.getPath() + " does not exist.");
            }
        } catch (Exception e) {
            e.printStackTrace();
            result = false;
        }
        return result;
    }

    /**
     * processing template.
     * 
     * @param template
     *            path to template.
     * @param vars
     *            sequence of template vars.
     * @param saveTo
     *            path to location where report is generated.
     * @return true if generation successful, false - otherwise.
     */
    public boolean processTemplate(String template, Map<String, Object> vars,
            String saveTo) {
        boolean result = true;
        Configuration cfg = new Configuration();
        try {
            String fileSeparator = System.getProperty("file.separator");
            template = template.replace("/", fileSeparator);
            saveTo = saveTo.replace("\\", fileSeparator);
            // Выходной поток
            OutputStream os = new FileOutputStream(new File(saveTo));
            Writer out = new OutputStreamWriter(os,"UTF-8");
            // Шаблон
            File templFile = new File(template);
            cfg.setDirectoryForTemplateLoading(templFile.getAbsoluteFile()
                    .getParentFile());
            cfg.setObjectWrapper(new DefaultObjectWrapper());

            Template temp = cfg.getTemplate(templFile.getName(),"UTF-8");
            temp.process(vars, out);
            out.flush();
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
            result = false;
        }
        return result;
    }

    private Map<String, Object> processData(String pathToDataFile)
            throws FileNotFoundException {

        if (pathToDataFile == null || "".equals(pathToDataFile)) {
            Map<String, Object> result = new HashMap<String, Object>();
            result.put("reqList", new HashMap<String, Object>());
            return result;
        } else if (!new File(pathToDataFile).exists()) {
            throw new FileNotFoundException("File: " + pathToDataFile
                    + "doesn't exist.");
        }

        File data = new File(pathToDataFile);
        Map<String, Object> resultMap = new HashMap<String, Object>();

        try {
            Stack<Map<String, Object>> stack = new Stack<Map<String, Object>>();
            stack.push(resultMap);
            XMLInputFactory factory = XMLInputFactory.newInstance();
            Reader fReader = new InputStreamReader(new FileInputStream(data.getPath()),"UTF-8");
            XMLEventReader eReader = factory.createXMLEventReader(fReader);
            while (eReader.hasNext()) {
                XMLEvent event = eReader.nextEvent();
                processDataEvent(event, stack);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return resultMap;
    }

    private void processDataEvent(XMLEvent xmlr,
            Stack<Map<String, Object>> stack) {
        switch (xmlr.getEventType()) {
        case XMLStreamConstants.START_ELEMENT:
            processDataStart(xmlr.asStartElement(), stack);
            break;
        case XMLStreamConstants.CHARACTERS:
            processDataChar(xmlr.asCharacters(), stack);
            break;
        case XMLStreamConstants.END_ELEMENT:
            processDataEnd(xmlr.asEndElement(), stack);
            break;
        }
    }

    private void processDataStart(StartElement xmlElem,
            Stack<Map<String, Object>> stack) {
        String elemName = xmlElem.getName().getLocalPart().toLowerCase();
        if (DATA_OBJECT.equals(elemName)) {
            javax.xml.stream.events.Attribute attr;
            attr = xmlElem.getAttributeByName(new QName(DATA_NAME));
            if (attr != null) {
                String name = attr.getValue().replace(" ", "");
                attr = xmlElem.getAttributeByName(new QName(DATA_VALUE));
                if (attr != null) {
                    String value = attr.getValue();
                    Map<String, Object> map = stack.peek();
                    map.put(name, value);
                    stack.push(new HashMap<String, Object>());
                } else {
                    Map<String, Object> value = new HashMap<String, Object>();
                    Map<String, Object> map = stack.peek();
                    map.put(name, value);
                    stack.push(value);
                }
            } else {
                stack.push(new HashMap<String, Object>());
            }
        }
    }

    private void processDataEnd(EndElement xmlElem,
            Stack<Map<String, Object>> stack) {
        String elemName = xmlElem.getName().getLocalPart().toLowerCase();
        if (DATA_OBJECT.equals(elemName)) {
            stack.pop();
        }
    }

    private void processDataChar(Characters xmlElem,
            Stack<Map<String, Object>> stack) {
    }

    protected static void log(String mes) {
        System.out.println(mes != null ? mes : "");
    }
}
